#include <linux/module.h>
#include <linux/fs.h>		/* for file_operations */

#define DEFAULT_MAJOR	121
#define RTC_PORT(x)     (unsigned short)(0x70 + (x))
#define RTC_SECONDS	0
#define RTC_MINUTES	2
#define RTC_HOURS	4

MODULE_AUTHOR("Guto");
MODULE_DESCRIPTION("Just a dummy RTC module to play with");

/* MODULE PARAMETERS */
static int dev = DEFAULT_MAJOR;
//MODULE_PARM(dev,"i");  // deprecated
module_param(dev,int,0);
MODULE_PARM_DESC(dev,"Major device number");

/* PROTOTYPES */
static int rtc_ioctl(struct inode * inode, struct file * file,
		     unsigned int cmd, unsigned long arg);
static unsigned char get_rtc(unsigned char addr);

/* FILE OPERATIONS */
static struct file_operations rtc_fops = {
    owner:    NULL,
    llseek:   NULL,
    read:     NULL,
    write:    NULL,
    readdir:  NULL,
    poll:     NULL,
    ioctl:    &rtc_ioctl,
    mmap:     NULL,
    open:     NULL,
    flush:    NULL,
    release:  NULL,
    fsync:    NULL,
    fasync:   NULL,
    lock:     NULL,
//    readv:    NULL,   // deprecated
//    writev:   NULL,   // deprecated
    sendpage: NULL,
    get_unmapped_area: NULL
};

int init_module(void)
{
    printk("Loading module rtc with major=%d\n", dev);

    /* Register the device driver */
    return register_chrdev(dev, "rtc", &rtc_fops);
}

void cleanup_module(void)
{
    printk("Unloading module rtc\n");

    /* Unregister the device driver */
    unregister_chrdev(dev, "rtc");
}

static int rtc_ioctl(struct inode * inode, struct file * file,
		     unsigned int cmd, unsigned long arg)
{
    int h, m, s;

    /* Get the current time from RTC */
    s = get_rtc(RTC_SECONDS);
    m = get_rtc(RTC_MINUTES);
    h = get_rtc(RTC_HOURS);
    *((int *)arg) = (h << 16) | (m << 8) | s;

    printk("Current time is %x\n", *((int *)arg));

    return 0;
}

unsigned char get_rtc(unsigned char addr)
{
    unsigned char val;

    asm("outb %1,%0" : : "d"(RTC_PORT(0)),"a"(addr));
    asm("inb %1,%0" : "=a"(val) : "d"(RTC_PORT(1))); 

    return val;
}
